if 'backUpVals' not in globals():
    globals()['backUpVals'] = {}
if 'restoreMethods' not in globals():
    globals()['restoreMethods'] = {}

from .readMatrix import readMatrix
from .writeMatrix import writeMatrix, restoreMatrix
from .caput import caput, restoreChannel
from .writeChannels import writeChannels
from .restoreEpics import restoreEpics
from epics import caget



from epics import caget
import numpy as np


def readMatrix(basename, rows, cols, firstRow=1, firstCol=1, suffix=None):
    '''
    Reads epics channels to put matrix on a EPICS channel matrix.
    '''
    inpMat = np.zeros((rows, cols))
    for ii in range(firstRow, firstRow + rows):
        for jj in range(firstCol, firstCol + cols):
            chName = basename + '_' + str(ii) + '_' + str(jj)
            if suffix is not None:
                chName = chName + '_' + suffix
            inpMat[ii - firstRow, jj-firstCol] = caget(chName)
    return inpMat



from time import sleep
from epics import caget
from .caput import caput
from . import backUpVals, restoreMethods


def writeChannels(chanInfo, bak=backUpVals):
    '''
    Write epics channels given in a dictionary format together.
    '''
    # Allowing to write the value directly under channel names if no other
    # parameter is required.
    for ch in chanInfo['channels']:
        if not isinstance(chanInfo['channels'][ch], dict):
            chanInfo['channels'][ch] = {'value': chanInfo['channels'][ch]}
        else:
            if 'value' not in chanInfo['channels'][ch]:
                raise RuntimeError('No value found for ' + ch)

    discardBak = {}  # caput used in this function do not need to save backUps
    kwargs = {'bak': discardBak}
    for ch in chanInfo['channels']:
        for arg in ['wait', 'timeout']:
            kwargs[ch] = {}
            if arg in chanInfo:  # If present as global argument
                kwargs[ch][arg] = chanInfo[arg]
            elif arg in chanInfo['channels'][ch]:
                kwargs[ch][arg] = chanInfo['channels'][ch][arg]

    curVals = {}
    for arg in ['tramp', 'wait', 'timeout']:
        if arg in chanInfo:
            curVals[arg] = chanInfo[arg]
    curVals['channels'] = {}
    for ch in chanInfo['channels']:
        curVals['channels'][ch] = {}
        for key in chanInfo['channels'][ch]:
            if key == 'value':
                curVals['channels'][ch][key] = caget(ch)
            else:
                curVals['channels'][ch][key] = chanInfo['channels'][ch][key]
    if all([set(curVals['channels']) != set(ele['value']['channels'])
            for ele in bak.values() if ele['type'] == 'group']):
        name = 'group_of|' + '|'.join(list(curVals['channels'].keys()))
        bak[name] = {'type': 'group', 'value': curVals, 'sno': len(bak)}

    rampYourself = True
    if 'tramp' in chanInfo:
        oldTramps = {}
        if chanInfo['tramp'] > 0:
            for ch in chanInfo['channels']:
                fus = ch.rfind('_') + 1
                if ch[fus:] in ['GAIN', 'OFFSET']:
                    try:
                        trampCh = ch[:fus] + 'TRAMP'
                        oldTramps[trampCh] = caget(trampCh)
                        rampYourself = rampYourself and False
                    except BaseException:
                        rampYourself = rampYourself or True
                else:
                    rampYourself = rampYourself or True
            if rampYourself:
                for ch in oldTramps:
                    caput(ch, 0, bak=discardBak)
                rampSteps = 100
                sleepTime = chanInfo['tramp'] / rampSteps
                curVals = {}
                for ch in chanInfo['channels']:
                    curVals[ch] = caget(ch)
                stepVals = {}
                for ch in curVals:
                    stepVals[ch] = (chanInfo['channels'][ch]['value']
                                    - curVals[ch]) / rampSteps
                for tstep in range(1, rampSteps):
                    sleep(sleepTime)
                    for ch in chanInfo['channels']:
                        caput(ch, curVals[ch] + tstep * stepVals[ch],
                              **kwargs[ch])
            else:
                for ch in oldTramps:
                    caput(ch, chanInfo['tramp'], bak=discardBak)
        # Write final values
        for ch in chanInfo['channels']:
            caput(ch, chanInfo['channels'][ch]['value'], **kwargs[ch])
        if not rampYourself:
            sleep(chanInfo['tramp'])
        sleep(0.1)  # Buffer time
        for ch in oldTramps:
            caput(ch, oldTramps[ch], bak=discardBak)
    else:
        for ch in chanInfo['channels']:
            if 'tramp' in chanInfo['channels'][ch]:
                kwargs[ch]['tramp'] = chanInfo['channels'][ch]['tramp']
            caput(ch, chanInfo['channels'][ch]['value'], **kwargs[ch])


def restoreGroup(bakVal):
    writeChannels(chanInfo=bakVal['value'], bak={})


restoreMethods['group'] = restoreGroup



from epics import caput
import numpy as np
from .readMatrix import readMatrix
from time import sleep
from . import backUpVals, restoreMethods


def writeMatrix(basename, inpMat, firstRow=1, firstCol=1, suffix=None,
                tramp=3.0, bak=backUpVals):
    '''
    Write epics channels to put matrix on a EPICS channel matrix.
    '''
    curMat = readMatrix(basename=basename, rows=inpMat.shape[0],
                        cols=inpMat.shape[1], firstRow=firstRow,
                        firstCol=firstCol, suffix=suffix)
    if basename not in bak:
        bak[basename] = {'type': 'matrix', 'name': basename, 'suffix': suffix,
                         'value': curMat, 'firstRow': firstRow,
                         'firstCol': firstCol, 'tramp': tramp, 'sno': len(bak)}
    # Special case, changing filter gains, can use TRAMP
    rampYourself = True
    if suffix == 'GAIN':
        try:
            curTramp = readMatrix(basename=basename, rows=inpMat.shape[0],
                                  cols=inpMat.shape[1], firstRow=firstRow,
                                  firstCol=firstCol, suffix='TRAMP')
            for ii in range(firstRow, firstRow + np.shape(inpMat)[0]):
                for jj in range(firstCol, firstCol + np.shape(inpMat)[1]):
                    chName = (basename + '_' + str(ii) + '_' + str(jj)
                              + '_TRAMP')
                    caput(chName, tramp)
            rampYourself = False
        except BaseException:
            rampYourself = True
    if rampYourself:
        rampSteps = 100
        # Get current matrix values
        stepMat = (inpMat - curMat) / rampSteps
        for tstep in range(1, rampSteps):
            for ii in range(firstRow, firstRow + np.shape(inpMat)[0]):
                for jj in range(firstCol, firstCol + np.shape(inpMat)[1]):
                    chName = basename + '_' + str(ii) + '_' + str(jj)
                    if suffix is not None:
                        chName = chName + '_' + suffix
                    matToWrite = (curMat[ii-firstRow, jj-firstCol]
                                  + tstep * stepMat[ii-firstRow, jj-firstCol])
                    caput(chName, matToWrite)
            sleep(tramp/rampSteps)
    # Finally write the required matrix
    for ii in range(firstRow, firstRow + np.shape(inpMat)[0]):
        for jj in range(firstCol, firstCol + np.shape(inpMat)[1]):
            chName = basename + '_' + str(ii) + '_' + str(jj)
            if suffix is not None:
                chName = chName + '_' + suffix
            caput(chName, inpMat[ii-firstRow, jj-firstCol])
    if not rampYourself:
        sleep(tramp + 0.5)    # Wait for ramping to end
        for ii in range(firstRow, firstRow + np.shape(inpMat)[0]):
            for jj in range(firstCol, firstCol + np.shape(inpMat)[1]):
                chName = (basename + '_' + str(ii) + '_' + str(jj)
                          + '_TRAMP')
                caput(chName, curTramp[ii-firstRow, jj-firstCol])


def restoreMatrix(bakVal):
    writeMatrix(basename=bakVal['name'], suffix=bakVal['suffix'],
                inpMat=bakVal['value'], firstCol=bakVal['firstCol'],
                firstRow=bakVal['firstRow'], tramp=bakVal['tramp'], bak={})


restoreMethods['matrix'] = restoreMatrix



import epics
from . import backUpVals, restoreMethods
from time import sleep


def caput(pvname, value, wait=False, timeout=60.0, bak=backUpVals, tramp=None):
    val = epics.caget(pvname, timeout=timeout)
    if pvname not in bak:
        bak[pvname] = {'type': 'channel', 'value': val, 'tramp': tramp,
                       'sno': len(bak), 'name': pvname}
    if tramp is None:
        epics.caput(pvname=pvname, value=value, wait=wait, timeout=timeout)
    else:
        rampYourself = True
        # If this looks like a GAIN channel, look for TRAMP channel
        if pvname[-4:] == 'GAIN':
            try:
                trch = pvname[:-4] + 'TRAMP'
                tbak = epics.caget(pvname=trch, timeout=timeout)
                rampYourself = False
            except BaseException:
                rampYourself = True
        if not rampYourself:
            epics.caput(pvname=trch, value=tramp, wait=wait, timeout=timeout)
            epics.caput(pvname=pvname, value=value, wait=wait, timeout=timeout)
            sleep(tramp)
            epics.caput(pvname=trch, value=tbak, wait=wait, timeout=timeout)
        else:
            rampSteps = 100
            valStep = (value - val) / rampSteps
            for tstep in range(1, rampSteps):
                epics.caput(pvname=pvname, value=val + valStep * tstep,
                            wait=wait, timeout=timeout)
                sleep(tramp / rampSteps)
            epics.caput(pvname=pvname, value=value, wait=wait, timeout=timeout)


def restoreChannel(bakVal):
    epics.caput(bakVal['name'], bakVal['value'])


restoreMethods['channel'] = restoreChannel



from . import backUpVals, restoreMethods


def restoreEpics(bak=backUpVals):
    print('Restoring channel values...')
    for ii in range(len(bak)-1, -1, -1):
        for ele in bak.values():
            if ii == ele['sno']:
                restoreMethods[ele['type']](ele)



from setuptools import setup


with open("README.md", "r") as fh:
    long_description = fh.read()

version = '0.3.0.6'
short_des = ('A simple package that gives wrapped caput and writeMatrix '
             + 'functions for writing EPICS channels which save previous '
             + 'values and a restoreEpics function can be used later to '
             + 'restore all values in case of error, interrupt, or as a '
             + 'final restore.')
dwnld_url = ('https://gitlab.com/anchal-physics/restoreepics/-/archive/'
             + version + '/restoreepics-' + version + '.tar.gz')
# Chose either "3 - Alpha" or "4 - Beta"
# or "5 - Production/Stable" as the current state of your package
classifiers = ['Development Status :: 4 - Beta',
               'Intended Audience :: Developers',
               'Topic :: Software Development :: Build Tools',
               'License :: OSI Approved :: MIT License',
               'Programming Language :: Python :: 3']

setup(name='restoreEpics',
      packages=['restoreEpics'],
      version=version,
      license='LICENSE',
      description=short_des,
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Anchal Gupta',
      author_email='anchal@caltech.edu',
      url='https://gitlab.com/anchal-physics/restoreepics',
      download_url=dwnld_url,
      keywords=['EPICS', 'RESTORE', 'MATRIX'],
      install_requires=['pyepics', 'argparse', 'numpy', 'PyYAML'],
      classifiers=classifiers,
      scripts=['bin/readMatrix', 'bin/writeMatrix', 'bin/caputt'])
